#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "keyvalue.h"
#include "autoschedver.h"

/* Defaults */

int version = 1;
int epoch = -1;
char psrid = 0;
char expcode[20] = "bc123";
double year = 2000.0;
double lst;
struct KeyValue *kv = 0;
int np=0, ng=0, ni=0;
int maxpass = 3;
double arcminsep=6.0;

char psrname[100];
char pointname[100] = "point";
char calname[20][100];
char geodname[20][100];
int psrdwell, caldwell[20], geoddwell[20];
double obslen;  /* seconds */

char contactfile[1000] = "contact.in";
char corfile[1000]     = "corparams.in";
char ranfile[1000]     = "random.in";
char setupfile[1000]   = "setup.in";

void writetop(FILE *out)
{
	time_t t;

	t = time(0);

	fprintf(out, "! VLBA sched file for pulsar parallax project\n!\n");
	fprintf(out, "! This file was automatically generated by autosched Ver. %s\n", VERSION);
	fprintf(out, "! at %s!\n", ctime(&t));
	fprintf(out, "! Total observation length should be %f seconds\n\n", obslen);
	fprintf(out, "version = %d\n", version);
	fprintf(out, "overwrite\n\n");
	fprintf(out, "expcode = %s\n\n", expcode);
}

void dumpfile(FILE *out, const char *infile)
{
	FILE *in;
	char buffer[1000];

	in = fopen(infile, "r");
	if(!in)
	{
		fprintf(stderr, "Cannot open %s.  Skipping\n", infile);
		return;
	}
	while(1)
	{
		fgets(buffer, 999, in);
		if(feof(in)) break;
		fprintf(out, buffer);
	}
	fclose(in);
}

void writecontact(FILE *out)
{
	fprintf(out, "! Contact\n\n");
	dumpfile(out, contactfile);
	fprintf(out, "\n");
}

void writecorrelation(FILE *out)
{
	int i, max;
	char key[100];

	max = ni;
	if(max > maxpass-1) max = maxpass-1;
	
	fprintf(out, "! Correlation\n\n");
	dumpfile(out, corfile);
	if(ni > 0)
	{
		fprintf(out, "cornote1= \'Multiple cor. passes requested for source %s\'\n", pointname);
		fprintf(out, "cornote2= \'Pass 1 : Use coords for source %s and use pulsar gate\'\n",
			psrname);
		for(i = 0; i < max; i++)
		{
			sprintf(key, "ib_cal_name%d", i+1);
			fprintf(out, "cornote%d= \'Pass %d : Use coords for source %s without pulsar gate\'\n", i+3, i+2, getKeyValuestring(kv, key));
		}
	}
	else
	{
		fprintf(out, "cornote1= \'Please gate source %s at correlation time.\'\n", psrname);
	}
	fprintf(out, "\n");
}

double optdate()
{
	int i, j;
	double a[4];
	double min = 1000.0, y = 0.0;
	a[0] = getKeyValuedouble(kv, "pi_max1");
	a[1] = getKeyValuedouble(kv, "pi_max2");
	a[2] = getKeyValuedouble(kv, "pi_min1");
	a[3] = getKeyValuedouble(kv, "pi_min2");

	for(i = 0; i < 4; i++) for(j = -10; j <= 10; j++)
	{
		if(fabs(a[i]+j-year) < min)
		{
			min = fabs(a[i]+j-year);
			y = a[i]+j;
		}
	}
	
	if(min > 0.49)
		fprintf(stderr, "Something fishy about optimal date!\n");

	return y;
}

void writerandom(FILE *out)
{
	int iy, id;
        double y, days = 365.00;
        int im = 0;
        int dm[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
        char sm[12][4] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
                          "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
	
	fprintf(out, "! Random\n\n");
	dumpfile(out, ranfile);
	y = optdate();
	iy = (int)y;
        if(iy % 4 == 0)
        {
                days = 366.00;
                dm[1]++;
        }
        id = (int)((y-iy)*days)+1;
        while(id > dm[im]) id -= dm[im++];
	
	fprintf(out, "note1   = \'The optimal observation date is %s %d.\'\n", sm[im], id);
	fprintf(out, "note2   = \'Please try to observe within 2 weeks of this date.\'\n");

	fprintf(out, "\n");
}

void writesetup(FILE *out)
{
	fprintf(out, "! Setup\n\n");
	dumpfile(out, setupfile);
	fprintf(out, "\n");
}

void writeobstime(FILE *out)
{
	int h, m, day;

	h = lst;
	m = 60.0*(lst-h);
	
	day = 52395 + (year-2002.33)*365.25;
	
	fprintf(out, "! When\n\n");
	fprintf(out, "day   = %d   ! year %f\n", day, year);
	fprintf(out, "LST   = \'VLBA_PT\'\n");
	fprintf(out, "start = %02d:%02d:00\n\n\n", h, m);
}


double text2double(const char *str)
{
	int colons = 0;
	int len;
	double sign = 1.0;
	char str2[100], *str3;
	double h = 0, m, s;
	double ret = -999999.999;
	
	for(len = 0; (str2[len] = str[len]) != 0; len++) 
		if(str[len] == ':') colons++;

	if(str[0] == '-' || str[0] == '+')
	{
		if(str[0] == '-') sign = -1.0;
		str3 = str2+1;
	}
	else str3 = str2;

	if(colons == 0) 
	{
		int digs;
		for(digs = 0; str3[digs] != 0 && str3[digs] != '.'; digs++);
		switch(digs)
		{
			case 0:
			case 1:
			case 2:
				if(sscanf(str, "%lf", &h) == 1) ret = h;
				break;
			case 4:
				h = 10.0*(str3[0] - '0') + (str3[1] - '0');
				if(sscanf(str3 + 2, "%lf", &m) == 1)
					ret = sign*(h + m/60.0);
				break;
			case 7:
				h = 100.0*(str3[0] - '0');
				str3++;
			case 6:
				h += 10.0*(str3[0] - '0') + (str3[1] - '0');
				m =  10.0*(str3[2] - '0') + (str3[3] - '0');
				if(sscanf(str3+4, "%lf", &s) == 1)
					ret = sign*(h + m/60.0 + s/3600.0);
				break;
		}
	}
	else if(colons == 2)
	{
		if(sscanf(str, "%lf:%lf:%lf", &h, &m, &s) == 3) 
		{
			ret = fabs(h) + m/60.0 + s/3600.0;
			if(h < 0) ret = -ret;
		}
	}

	return ret;
}

void double2text(double num, char *str, char sep)
{
	int h, m;

	if(num < 0)
	{
		num = -num;
		str[0] = '-';
		str++;
	}
	h = (int)num;
	num -= h;
	num *= 60.0;
	m = (int)num;
	num -= m;
	num *= 60.0;
	sprintf(str, "%02d%c%02d%c%09.6f", h, sep, m, sep, num);
}

void writepointing(FILE *out, const char *psrrastr, const char *psrdecstr)
{
	double psrra, psrdec, r, d;
	double dx=0.0, dy=0.0, R, cosd;
	int i;
	char key[100];
	double maxsep;
	static int first = 1;
	
	maxsep = arcminsep*M_PI/(180.0*60.0);  

	if(ni < 1)
	{
		strcpy(pointname, psrname);
		return;
	}

	psrra = (M_PI/12.0)*text2double(psrrastr);
	psrdec = (M_PI/180.0)*text2double(psrdecstr);

	cosd = cos(psrdec);

	for(i = 0; i < ni; i++)
	{
		sprintf(key, "ib_cal_ra%d", i+1);
		r = (M_PI/12.0)*text2double(getKeyValuestring(kv, key));
		sprintf(key, "ib_cal_dec%d", i+1);
		d = (M_PI/180.0)*text2double(getKeyValuestring(kv, key));

		dx += (r-psrra)*cosd;
		dy += d-psrdec;
	}

	dx /= (ni+1.0);
	dy /= (ni+1.0);

	R = sqrt(dx*dx + dy*dy)/maxsep;
	if(R > 1.0)
	{
		dx /= R;
		dy /= R;
		if(first) fprintf(stderr, "pointing center shift wants %f arcmin,"
			" more than limit of %f\n", arcminsep*R, arcminsep);
	}

	d = (180.0/M_PI)*(dy+psrdec);
	r = (12.0/M_PI)*(dx/cosd + psrra);

	/* get truncated name */
	strcpy(key, psrname);
	key[8] = 0;
	sprintf(pointname, "%s.pt", key);

	fprintf(out, "  ! pointing center\n");
	fprintf(out, "    source = \'%s\'", pointname);
	double2text(r, key, ':');
	fprintf(out, "  ra = %s", key);
	double2text(d, key, ':');
	fprintf(out, "  dec = %s  epoch = \'J2000\' /\n\n", key);
	first = 0;
}

void writesources(FILE *out, const char *xfile)
{
	FILE *pin;
	char command[1000];
	char response[1000];
	char junk[100], ra[100], dec[100], key[100];
	int i;
	
	fprintf(out, "! Sources\n\n");
	fprintf(out, "srccat /\n\n");
	fprintf(out, "  ! Pulsar  epoch = %f\n", year);
	
	sprintf(command, "pmpar %s -z %f", xfile, year);
	pin = popen(command, "r");
	fgets(response, 999, pin);
	fclose(pin);
//	fprintf(stderr, "response from pmpar: %s\n", response);

	sscanf(response, "%s %s %s", junk, ra, dec);
	strcpy(psrname, getKeyValuestring(kv, "name"));
	psrdwell = getKeyValueint(kv, "psr_dwell");
	fprintf(out, "    source = \'%s\'  ra = %s  dec = %s  epoch = \'J2000\'  calcode = 'G' /\n\n", 
		psrname, ra, dec);
	writepointing(out, ra, dec);

	if(np > 0) 
	{
		fprintf(out, "  ! Phase calibrators\n");
		for(i = 0; i < np; i++)
		{
			sprintf(key, "phase_cal_name%d", i+1);
			strcpy(calname[i], getKeyValuestring(kv, key));
			sprintf(key, "phase_cal_dwell%d", i+1);
			caldwell[i] = getKeyValueint(kv, key);
			sprintf(key, "phase_cal_sep%d", i+1);
			fprintf(out, "    ! sep = %s arcmin\n", getKeyValuestring(kv, key));
			fprintf(out, "    source = \'%s\'", calname[i]);
			sprintf(key, "phase_cal_ra%d", i+1);
			fprintf(out, "  ra = %s", getKeyValuestring(kv, key));
			sprintf(key, "phase_cal_dec%d", i+1);
			fprintf(out, "  dec = %s  epoch = \'J2000\' /\n\n", getKeyValuestring(kv, key));
		}
	}
	
	if(ng > 0) 
	{
		fprintf(out, "  ! Geodesy calibrators\n");
		for(i = 0; i < ng; i++)
		{
			sprintf(key, "geod_cal_name%d", i+1);
			strcpy(geodname[i], getKeyValuestring(kv, key));
			fprintf(out, "    source = \'%s\'", geodname[i]);
			sprintf(key, "geod_cal_dwell%d", i+1);
			geoddwell[i] = getKeyValueint(kv, key);
			sprintf(key, "geod_cal_ra%d", i+1);
			fprintf(out, "  ra = %s", getKeyValuestring(kv, key));
			sprintf(key, "geod_cal_dec%d", i+1);
			fprintf(out, "  dec = %s  epoch = \'J2000\' /\n\n", getKeyValuestring(kv, key));
		}
	}
	
	if(ni > 0) 
	{
		fprintf(out, "  ! In-beam calibrators\n");
		for(i = 0; i < ni; i++)
		{
			sprintf(key, "ib_cal_name%d", i+1);
			fprintf(out, "    source = \'%s\'", getKeyValuestring(kv, key));
			sprintf(key, "ib_cal_ra%d", i+1);
			fprintf(out, "  ra = %s", getKeyValuestring(kv, key));
			sprintf(key, "ib_cal_dec%d", i+1);
			fprintf(out, "  dec = %s  epoch = \'J2000\' /\n\n", getKeyValuestring(kv, key));
		}
	}

	fprintf(out, "endcat /\n\n\n");
}

void writegeod(FILE *out)
{
	int i;
	fprintf(out, "  ! geodesy sources\n");
	for(i = 0; i < ng; i++)
		fprintf(out, "  source = \'%s\' dwell = %d/\n", geodname[i],
			geoddwell[i]);
	fprintf(out, "\n");
}

void writescans(FILE *out, int dloops, int dt)
{
	int rep;
	double loop;
	int i;
	
	loop = psrdwell+30.0;
	for(i = 0; i < np; i++) loop += (caldwell[i]+30.0);
	rep = (obslen-360.0*ng)/loop + dloops;
	
	fprintf(out, "! Scans\n\n");

	/* first round of geodesy sources */
	writegeod(out);

	fprintf(out, "  ! readback tests\n");
	fprintf(out, "  gap = 2:00 /\n");
	fprintf(out, "  gap = 0:00 /\n\n");

	fprintf(out, "  ! Main loop\n");
	fprintf(out, "  source = \'%s\' dwell = %d /\n", calname[0], caldwell[0] + dt);
	fprintf(out, "  group %d rep %d\n", np+1, rep);
	fprintf(out, "    source = \'%s\' dwell = %d /\n", pointname, psrdwell);
	for(i = np-1; i >= 0; i--)
		fprintf(out, "    source = \'%s\' dwell = %d /\n", calname[i], caldwell[i]);
	fprintf(out, "\n");

	/* last round of geodesy sources */
	writegeod(out);
}

double runsched(const char *filename)
{
	char command[1000], line[1000];
	FILE *pin;
	double lapse = 0;
	int i, h1, h2, m1, m2, s1, s2;
	
	sprintf(command, "sched < %s", filename);
	
	pin = popen(command, "r");
	for(;;)
	{
		fgets(line, 999, pin);
		if(feof(pin)) break;
		line[22] = 0;
		if(strcmp(line, "SCHOPT:  There will be") == 0)
		{
			for(i = 23; line[i]; i++) if(line[i] == '/') break;
			if(line[i] != '/') 
			{
				fclose(pin);
				return -1;
			}
			sscanf(line+i+1, "%d:%d:%d", &h1, &m1, &s1);
			for(i++; line[i]; i++) if(line[i] == '/') break;
			if(line[i] != '/')
			{
				fclose(pin);
				return -1;
			}
			sscanf(line+i+1, "%d:%d:%d", &h2, &m2, &s2);

			if(h2 < h1) h2 += 24;
	
			lapse = 3600*(h2-h1) + 60*(m2-m1) + s2-s1;
		}
	}
	fclose(pin);

	return lapse;
}

void writeschedfile(const char *outfile, const char *xfile, int dloops, int dt)
{
	FILE *out;
	
	if(strcmp(outfile, "-") != 0) out = fopen(outfile, "w");
	else out = stdout;

	writetop(out);
	writecontact(out);
	writecorrelation(out);
	writerandom(out);
	writesetup(out);
	writeobstime(out);
	writesources(out, xfile);
	writescans(out, dloops, dt),
	fprintf(out, "exit\n");

	if(strcmp(outfile, "-") != 0) fclose(out);
}

void makefile(const char *outfile, const char *xfile)
{
	double t1, t2, t3, t4, t5;
	int dloops, dt;

	writeschedfile(outfile, xfile, -1, 0);
	t1 = runsched(outfile);
	writeschedfile(outfile, xfile, 0, 0);
	t2 = runsched(outfile);
	
	dloops = (obslen-t2)/(t2-t1);
	if(obslen < t2) dloops--;

	writeschedfile(outfile, xfile, dloops, 0);
	t3 = runsched(outfile);

	dt = obslen-t3;

	writeschedfile(outfile, xfile, dloops, dt);
	t4 = runsched(outfile);

	dt += obslen-t4;
	writeschedfile(outfile, xfile, dloops, dt);
	t5 = runsched(outfile);

	fprintf(stderr, "%f %f %f %f %f\n", t1, t2, t3, t4, t5);
}

int main(int argc, char **argv)
{
	char outfile[1000] = "-", key[100];
	int i, j;
	double d;
	char k[100], v[100];

	if(argc < 5)
	{
		fprintf(stderr, "VERSION %s\n\n", VERSION);
		fprintf(stderr, "Usage : %s <.x file> <year> <lst> <expcode> [.x overrides]\n", argv[0]);
		fprintf(stderr, "\n  year is to determine pulsar position\n");
		fprintf(stderr, "  lst is start LST of observation\n");
		fprintf(stderr, "  [.x overrides] are any overrides to the .x file, such as psr_dwell=99\n");
		return 0;
	}

	year = atof(argv[2]);
	if(year < 1990 || year > 2100)
	{
		fprintf(stderr, "year parameter (%s) needs to be > 1990 and < 2100\n", argv[2]);
		return 0;
	}
	lst = atof(argv[3]);
	if(lst < 0 || lst > 24)
	{
		fprintf(stderr, "lst parameter (%s) needs to be >= 0 and <= 24\n", argv[3]);
		return 0;
	}
	
	strcpy(expcode, argv[4]);
	
	sprintf(outfile, "%s.key", expcode);
	
	kv = loadKeyValue(argv[1]);
	if(!kv)
	{
		fprintf(stderr, "can't read %s.\n", argv[1]);
		return 0;
	}

	if(argc > 5) for(i = 5; i < argc; i++)
        {
                if(argv[i][0] == '=') fprintf(stderr,
                        "Warning -- option begins with =\n");
                for(j = 0; argv[i][j] != 0; j++) if(argv[i][j] == '=')
                        argv[i][j] = ' ';
                if(sscanf(argv[i], "%s %s", k, v) != 2) continue;

                j = KeyValuekeyindex(kv, k);
                if(j < 0) KeyValueaddparm(kv, k, v);
                else
                {
                        free(kv->value[j]);
                        kv->value[j] = strdup(v);
                }
        }

	i = getKeyValueint(kv, "max_pass");
	if(i != KV_INTERR) maxpass = i;

	d = getKeyValuedouble(kv, "max_offset");
	if(d != KV_FLOATERR) arcminsep = d;

	obslen = getKeyValuedouble(kv, "obs_length");

	for(np = 0; np < 20; np++)
	{
		sprintf(key, "phase_cal_name%d", np+1);
		if(getKeyValuestring(kv, key) == 0) break;
	}
	
	for(ng = 0; ng < 20; ng++)
	{
		sprintf(key, "geod_cal_name%d", ng+1);
		if(getKeyValuestring(kv, key) == 0) break;
	}

	for(ni = 0; ni < 20; ni++)
	{
		sprintf(key, "ib_cal_name%d", ni+1);
		if(getKeyValuestring(kv, key) == 0) break;
	}
	
	fprintf(stderr, "Calibrator count : %d %d %d\n", np, ng, ni);
	
	makefile(outfile, argv[1]);

	return 0;
}
